"""
Implements the connection to the ihc controller
"""

# pylint: disable=bare-except
import base64
import datetime
import io
import zlib
from ihcsdk.ihcconnection import IHCConnection
from ihcsdk.ihcsslconnection import IHCSSLConnection

IHCSTATE_READY = "text.ctrl.state.ready"


class IHCSoapClient:
    """Implements a limited set of the soap request for the IHC controller"""

    ihcns = {
        "SOAP-ENV": "http://schemas.xmlsoap.org/soap/envelope/",
        "ns1": "utcs",
        "ns2": "utcs.values",
        "ns3": "utcs.values",
    }

    def __init__(self, url: str):
        """Initialize the IIHCSoapClient with a url for the controller"""
        self.url = url
        self.username = ""
        self.password = ""
        if url.startswith("https://"):
            self.connection = IHCSSLConnection(url)
        else:
            self.connection = IHCConnection(url)

    def authenticate(self, username: str, password: str) -> bool:
        """Do an Authentricate request and save the cookie returned to be used
        on the following requests.
        Return True if the request was successfull
        """
        self.username = username
        self.password = password

        auth_payload = """<authenticate1 xmlns=\"utcs\"
                          xmlns:i=\"http://www.w3.org/2001/XMLSchema-instance\">
                          <password>{password}</password>
                          <username>{username}</username>
                          <application>treeview</application>
                          </authenticate1>"""
        payload = auth_payload.format(password=self.password, username=self.username)

        xdoc = self.connection.soap_action(
            "/ws/AuthenticationService", "authenticate", payload
        )
        if xdoc is not False:
            isok = xdoc.find(
                "./SOAP-ENV:Body/ns1:authenticate2/ns1:loginWasSuccessful",
                IHCSoapClient.ihcns,
            )
            return isok.text == "true"
        return False

    def get_state(self) -> str:
        """Get the controller state"""
        xdoc = self.connection.soap_action("/ws/ControllerService", "getState", "")
        if xdoc is not False:
            return xdoc.find(
                "./SOAP-ENV:Body/ns1:getState1/ns1:state", IHCSoapClient.ihcns
            ).text
        return False

    def wait_for_state_change(self, state: str, waitsec) -> str:
        """Wait for controller state change and return state"""
        payload = """<ns1:waitForControllerStateChange1
                     xmlns:ns1=\"utcs\" xsi:type=\"ns1:WSControllerState\">
                     <ns1:state xsi:type=\"xsd:string\">{state}</ns1:state>
                     </ns1:waitForControllerStateChange1>
                     <ns2:waitForControllerStateChange2
                     xmlns:ns2=\"utcs\" xsi:type=\"xsd:int\">
                     {wait}</ns2:waitForControllerStateChange2>
                     """.format(
            state=state, wait=waitsec
        )
        xdoc = self.connection.soap_action(
            "/ws/ControllerService", "waitForControllerStateChange", payload
        )
        if xdoc is not False:
            return xdoc.find(
                "./SOAP-ENV:Body/ns1:waitForControllerStateChange3/ns1:state",
                IHCSoapClient.ihcns,
            ).text
        return False

    def get_project(self) -> str:
        """Get the ihc project in single SOAP action.
        You should use the get_project_in_segments to get the project in multiple segments.
        This will stress the IHC controller less
        """
        xdoc = self.connection.soap_action("/ws/ControllerService", "getIHCProject", "")
        if xdoc is not False:
            base64data = xdoc.find(
                "./SOAP-ENV:Body/ns1:getIHCProject1/ns1:data", IHCSoapClient.ihcns
            ).text
            if not base64:
                return False
            compresseddata = base64.b64decode(base64data)
            return zlib.decompress(compresseddata, 16 + zlib.MAX_WBITS).decode(
                "ISO-8859-1"
            )
        return False

    def get_project_in_segments(self, info=None) -> str:
        """Get the ihc project per segments.
        Param: info .. reuse existing project info. If not provided, the get_project_info() is called internally.
        """
        if info is None:
            info = self.get_project_info()
        if info:
            projectMajor = info.get("projectMajorRevision", 0)
            projectMinor = info.get("projectMinorRevision", 0)
            buffer = io.BytesIO()
            for s in range(self.get_project_number_of_segments()):
                segment = self.get_project_segment(s, projectMajor, projectMinor)
                if segment is False:
                    return False
                buffer.write(segment)
            return zlib.decompress(buffer.getvalue(), 16 + zlib.MAX_WBITS).decode(
                "ISO-8859-1"
            )
        return False

    def get_project_info(self) -> dict:
        """Returns dictionary of project info items."""
        xdoc = self.connection.soap_action(
            "/ws/ControllerService", "getProjectInfo", ""
        )
        if xdoc is not False:
            info = {}
            elem = xdoc.find("./SOAP-ENV:Body/ns1:getProjectInfo1", IHCSoapClient.ihcns)
            if elem is not None:
                for e in list(elem):
                    name = e.tag.split("}")[-1]
                    info[name] = IHCSoapClient.__get_value(e)
            return info
        return False

    def get_project_number_of_segments(self) -> int:
        """Returns the number of segments needed to fetch the current ihc-project."""
        xdoc = self.connection.soap_action(
            "/ws/ControllerService", "getIHCProjectNumberOfSegments", ""
        )
        if xdoc is not False:
            return int(
                xdoc.find(
                    "./SOAP-ENV:Body/ns1:getIHCProjectNumberOfSegments1",
                    IHCSoapClient.ihcns,
                ).text
            )
        return False

    def get_project_segment(self, segment: int, projectMajor: int, projectMinor: int):
        """Returns a segment of the ihc-project with the given number.
        Returns null if the segment number increases above the number of segments available.
        The segments are offset from 0.
        The project-versions given as parameters are used to indentify the project that should be fetched.
        That is, to make sure that you suddenly don't get segments belonging to another project.
        """
        payload = """
            <getIHCProjectSegment1 xmlns="utcs">{segment}</getIHCProjectSegment1>
            <getIHCProjectSegment2 xmlns="utcs">{major}</getIHCProjectSegment2>
            <getIHCProjectSegment3 xmlns="utcs">{minor}</getIHCProjectSegment3>
            """.format(
            segment=segment, major=projectMajor, minor=projectMinor
        )
        xdoc = self.connection.soap_action(
            "/ws/ControllerService", "getIHCProjectSegment", payload
        )
        if xdoc is not False:
            base64data = xdoc.find(
                "./SOAP-ENV:Body/ns1:getIHCProjectSegment4/ns1:data",
                IHCSoapClient.ihcns,
            ).text
            if not base64:
                return False
            compresseddata = base64.b64decode(base64data)
            return compresseddata
        return False

    def set_runtime_value_bool(self, resourceid: int, value: bool) -> bool:
        """Set a boolean runtime value"""
        if value:
            boolvalue = "true"
        else:
            boolvalue = "false"

        payload = """
            <setResourceValue1 xmlns=\"utcs\"
            xmlns:i=\"http://www.w3.org/2001/XMLSchema-instance\">
            <value i:type=\"a:WSBooleanValue\" xmlns:a=\"utcs.values\">
            <a:value>{value}</a:value></value>
            <typeString/>
            <resourceID>{id}</resourceID>
            <isValueRuntime>true</isValueRuntime>
            </setResourceValue1>
            """.format(
            id=resourceid, value=boolvalue
        )
        xdoc = self.connection.soap_action(
            "/ws/ResourceInteractionService", "setResourceValue", payload
        )
        if xdoc is not False:
            result = xdoc.find(
                r"./SOAP-ENV:Body/ns1:setResourceValue2", IHCSoapClient.ihcns
            ).text
            return result == "true"
        return False

    def set_runtime_value_int(self, resourceid: int, intvalue: int):
        """Set a integer runtime value"""
        payload = """
            <setResourceValue1 xmlns=\"utcs\"
            xmlns:i=\"http://www.w3.org/2001/XMLSchema-instance\">
            <value i:type=\"a:WSIntegerValue\" xmlns:a=\"utcs.values\">
            <a:integer>{value}</a:integer></value>
            <typeString/>
            <resourceID>{id}</resourceID>
            <isValueRuntime>true</isValueRuntime>
            </setResourceValue1>
            """.format(
            id=resourceid, value=intvalue
        )
        xdoc = self.connection.soap_action(
            "/ws/ResourceInteractionService", "setResourceValue", payload
        )
        if xdoc is not False:
            result = xdoc.find(
                "./SOAP-ENV:Body/ns1:setResourceValue2", IHCSoapClient.ihcns
            ).text
            return result == "true"
        return False

    def set_runtime_value_float(self, resourceid: int, floatvalue: float):
        """Set a flot runtime value"""
        payload = """
            <setResourceValue1 xmlns=\"utcs\"
            xmlns:i=\"http://www.w3.org/2001/XMLSchema-instance\">
            <value i:type=\"a:WSFloatingPointValue\" xmlns:a=\"utcs.values\">
            <a:floatingPointValue>{value}</a:floatingPointValue></value>
            <typeString/>
            <resourceID>{id}</resourceID>
            <isValueRuntime>true</isValueRuntime>
            </setResourceValue1>
            </s:Body>
            """.format(
            id=resourceid, value=floatvalue
        )
        xdoc = self.connection.soap_action(
            "/ws/ResourceInteractionService", "setResourceValue", payload
        )
        if xdoc is not False:
            result = xdoc.find(
                "./SOAP-ENV:Body/ns1:setResourceValue2", IHCSoapClient.ihcns
            ).text
            return result == "true"
        return False

    def set_runtime_value_timer(self, resourceid: int, timer: int):
        """Set a timer runtime value in milliseconds"""
        payload = """
            <setResourceValue1 xmlns=\"utcs\"
            xmlns:i=\"http://www.w3.org/2001/XMLSchema-instance\">
            <value i:type=\"a:WSTimerValue\" xmlns:a=\"utcs.values\">
            <a:milliseconds>{value}</a:milliseconds></value>
            <typeString/>
            <resourceID>{id}</resourceID>
            <isValueRuntime>true</isValueRuntime>
            </setResourceValue1>
            </s:Body>
            """.format(
            id=resourceid, value=timer
        )
        xdoc = self.connection.soap_action(
            "/ws/ResourceInteractionService", "setResourceValue", payload
        )
        if xdoc is not False:
            result = xdoc.find(
                "./SOAP-ENV:Body/ns1:setResourceValue2", IHCSoapClient.ihcns
            ).text
            return result == "true"
        return False

    def set_runtime_value_time(
        self, resourceid: int, hours: int, minutes: int, seconds: int
    ):
        """Set a time runtime value in hours:minutes:seconds"""
        payload = f"""
            <setResourceValue1 xmlns=\"utcs\"
            xmlns:i=\"http://www.w3.org/2001/XMLSchema-instance\">
            <value i:type=\"a:WSTimeValue\" xmlns:a=\"utcs.values\">
            <a:hours>{hours}</a:hours>
            <a:minutes>{minutes}</a:minutes>
            <a:seconds>{seconds}</a:seconds>
            </value>
            <typeString/>
            <resourceID>{resourceid}</resourceID>
            <isValueRuntime>true</isValueRuntime>
            </setResourceValue1>
            </s:Body>
            """
        xdoc = self.connection.soap_action(
            "/ws/ResourceInteractionService", "setResourceValue", payload
        )
        if xdoc is not False:
            result = xdoc.find(
                "./SOAP-ENV:Body/ns1:setResourceValue2", IHCSoapClient.ihcns
            ).text
            return result == "true"
        return False

    def get_time(resource_value):

        hours = int(resource_value.find("./ns2:hours", IHCSoapClient.ihcns).text)
        minutes = int(resource_value.find("./ns2:minutes", IHCSoapClient.ihcns).text)
        seconds = int(resource_value.find("./ns2:seconds", IHCSoapClient.ihcns).text)

        return datetime.time(hours, minutes, seconds)

    def get_datetime(resource_value):
        year = int(resource_value.find("./ns1:year", IHCSoapClient.ihcns).text)
        month = int(
            resource_value.find("./ns1:monthWithJanuaryAsOne", IHCSoapClient.ihcns).text
        )
        day = int(resource_value.find("./ns1:day", IHCSoapClient.ihcns).text)
        hours = int(resource_value.find("./ns1:hours", IHCSoapClient.ihcns).text)
        minutes = int(resource_value.find("./ns1:minutes", IHCSoapClient.ihcns).text)
        seconds = int(resource_value.find("./ns1:seconds", IHCSoapClient.ihcns).text)
        return datetime.datetime(year, month, day, hours, minutes, seconds)

    def get_date(resource_value):
        year = int(resource_value.find("./ns2:year", IHCSoapClient.ihcns).text)
        if year == 0:
            year = datetime.datetime.today().year
        month = int(resource_value.find("./ns2:month", IHCSoapClient.ihcns).text)
        day = int(resource_value.find("./ns2:day", IHCSoapClient.ihcns).text)
        return datetime.datetime(year, month, day)

    def __get_value(resource_value):
        """Get a runtime value from the xml base on the type in the xml"""
        if resource_value is None:
            return None
        valuetype = resource_value.attrib[
            "{http://www.w3.org/2001/XMLSchema-instance}type"
        ].split(":")[1]
        result = resource_value.text
        match valuetype:
            case "WSBooleanValue":
                result = resource_value.find("./ns2:value", IHCSoapClient.ihcns).text == "true"
            case "WSIntegerValue":
                result = int( resource_value.find("./ns2:integer", IHCSoapClient.ihcns).text)
            case "WSFloatingPointValue":
                result = round(float(resource_value.find("./ns2:floatingPointValue", IHCSoapClient.ihcns).text), 2)
            case "WSEnumValue":
                result = resource_value.find("./ns2:enumName", IHCSoapClient.ihcns).text
            case "WSTimerValue":
                return int(resource_value.find("./ns2:milliseconds", IHCSoapClient.ihcns).text)
            case "WSTimeValue":
                result = IHCSoapClient.get_time(resource_value)
            case "WSDate":
                result = IHCSoapClient.get_datetime(resource_value)
            case "WSDateValue":
                result = IHCSoapClient.get_date(resource_value)
            case "int":
                result = int(resource_value.text)
        return result

    def get_runtime_value(self, resourceid: int):
        """Get runtime value of specified resource it
        The returned value will be boolean, integer or float
        Return None if resource cannot be found or on error
        """
        payload = """<getRuntimeValue1 xmlns="utcs">{id}</getRuntimeValue1>
                  """.format(
            id=resourceid
        )
        xdoc = self.connection.soap_action(
            "/ws/ResourceInteractionService", "getResourceValue", payload
        )
        if xdoc is False:
            return None
        value = xdoc.find(
            "./SOAP-ENV:Body/ns1:getRuntimeValue2/ns1:value", IHCSoapClient.ihcns
        )
        return IHCSoapClient.__get_value(value)

    def get_runtime_values(self, resourceids):
        """Get runtime values of specified resource ids
        Return None if resource cannot be found or on error
        """

        idsarr = ""
        for ihcid in resourceids:
            idsarr += "<arrayItem>{id}</arrayItem>".format(id=ihcid)
        payload = '<getRuntimeValues1 xmlns="utcs">' + idsarr + "</getRuntimeValues1>"
        xdoc = self.connection.soap_action(
            "/ws/ResourceInteractionService", "getResourceValues", payload
        )
        if xdoc is False:
            return False
        result = xdoc.findall(
            "./SOAP-ENV:Body/ns1:getRuntimeValues2/ns1:arrayItem", IHCSoapClient.ihcns
        )
        changes = {}
        for item in result:
            ihcid = item.find("ns1:resourceID", IHCSoapClient.ihcns)
            if ihcid is None:
                continue
            resourceValue = item.find("./ns1:value", IHCSoapClient.ihcns)
            itemValue = IHCSoapClient.__get_value(resourceValue)
            if itemValue is not None:
                changes[int(ihcid.text)] = itemValue
        return changes

    def cycle_bool_value(self, resourceid: int):
        """Turn a booelan resource On and back Off
        Return None if resource cannot be found or on error
        """
        setBool = (
            "<arrayItem>"
            '<value xsi:type="ns1:WSBooleanValue">'
            "<ns1:value>{value}</ns1:value>"
            "</value>"
            "<typeString></typeString>"
            "<resourceID>{id}</resourceID>"
            "<isValueRuntime>true</isValueRuntime>"
            "</arrayItem>"
        )
        payload = (
            '<setResourceValues1 xmlns="utcs" xmlns:ns1="utcs.values">'
            + setBool.format(value="true", id=resourceid)
            + setBool.format(value="false", id=resourceid)
            + "</setResourceValues1>"
        )
        xdoc = self.connection.soap_action(
            "/ws/ResourceInteractionService", "SOAPAction: setResourceValues", payload
        )
        if xdoc is False:
            return None
        return True

    def enable_runtime_notification(self, resourceid: int):
        """Enable notification for specified resource id"""
        return self.enable_runtime_notifications([resourceid])

    def enable_runtime_notifications(self, resourceids):
        """Enable notification for specified resource ids"""
        idsarr = ""
        for ihcid in resourceids:
            idsarr += "<a:arrayItem>{id}</a:arrayItem>".format(id=ihcid)

        payload = """<enableRuntimeValueNotifications1 xmlns=\"utcs\"
                     xmlns:a=\"http://www.w3.org/2001/XMLSchema\"
                     xmlns:i=\"http://www.w3.org/2001/XMLSchema-instance\">
                     {arr}
                     </enableRuntimeValueNotifications1>
                     """.format(
            arr=idsarr
        )
        xdoc = self.connection.soap_action(
            "/ws/ResourceInteractionService", "enableRuntimeValueNotifications", payload
        )
        return xdoc is not False

    def wait_for_resource_value_changes(self, wait: int = 10):
        """
        Long polling for changes and return a dictionary with resource:value
        for changes (Only last change)
        """
        change_list = self.wait_for_resource_value_change_list(wait)
        if change_list is False:
            return False
        last_changes = {}
        for id, value in change_list:
            last_changes[id] = value
        return last_changes

    def wait_for_resource_value_change_list(self, wait: int = 10):
        """
        Long polling for changes and return a resource id dictionary with a list of all changes since last poll.
        Return a list of tuples with the id,value
        """
        changes = []
        payload = """<waitForResourceValueChanges1
                     xmlns=\"utcs\">{timeout}</waitForResourceValueChanges1>
                  """.format(
            timeout=wait
        )
        xdoc = self.connection.soap_action(
            "/ws/ResourceInteractionService", "getResourceValue", payload
        )
        if xdoc is False:
            return False
        result = xdoc.findall(
            "./SOAP-ENV:Body/ns1:waitForResourceValueChanges2/ns1:arrayItem",
            IHCSoapClient.ihcns,
        )
        for item in result:
            ihcid = item.find("ns1:resourceID", IHCSoapClient.ihcns)
            if ihcid is None:
                continue

            resource_value = item.find("./ns1:value", IHCSoapClient.ihcns)
            value = IHCSoapClient.__get_value(resource_value)
            if value is not None:
                changes.append((int(ihcid.text), value))
        return changes

    def get_user_log(self, language="da"):
        """Get the controller state"""
        payload = """<getUserLog1 xmlns="utcs" />
                     <getUserLog2 xmlns="utcs">0</getUserLog2>
                     <getUserLog3 xmlns="utcs">{language}</getUserLog3>
                     """.format(
            language=language
        )
        xdoc = self.connection.soap_action(
            "/ws/ConfigurationService", "getUserLog", payload
        )
        if xdoc is not False:
            base64data = xdoc.find(
                "./SOAP-ENV:Body/ns1:getUserLog4/ns1:data", IHCSoapClient.ihcns
            ).text
            if not base64data:
                return False
            return base64.b64decode(base64data).decode("UTF-8")
        return False

    def clear_user_log(self):
        """Clear the user log in the controller"""
        self.connection.soap_action("/ws/ConfigurationService", "clearUserLog", "")
        return

    def get_system_info(self):
        """Get controller system info"""
        xdoc = self.connection.soap_action(
            "/ws/ConfigurationService", "getSystemInfo", ""
        )
        if xdoc is False:
            return False
        info = {
            "uptime": IHCSoapClient.__extract_sysinfo(xdoc, "uptime"),
            "realtimeclock": IHCSoapClient.__extract_sysinfo(xdoc, "realtimeclock"),
            "serial_number": IHCSoapClient.__extract_sysinfo(xdoc, "serialNumber"),
            "production_date": IHCSoapClient.__extract_sysinfo(xdoc, "productionDate"),
            "brand": IHCSoapClient.__extract_sysinfo(xdoc, "brand"),
            "version": IHCSoapClient.__extract_sysinfo(xdoc, "version"),
            "hw_revision": IHCSoapClient.__extract_sysinfo(xdoc, "hwRevision"),
            "sw_date": IHCSoapClient.__extract_sysinfo(xdoc, "swDate"),
            "dataline_version": IHCSoapClient.__extract_sysinfo(
                xdoc, "datalineVersion"
            ),
            "rf_module_software_version": IHCSoapClient.__extract_sysinfo(
                xdoc, "rfModuleSoftwareVersion"
            ),
            "rf_module_serial_number": IHCSoapClient.__extract_sysinfo(
                xdoc, "rfModuleSerialNumber"
            ),
            "application_is_without_viewer": IHCSoapClient.__extract_sysinfo(
                xdoc, "applicationIsWithoutViewer"
            ),
            "sms_modem_software_version": IHCSoapClient.__extract_sysinfo(
                xdoc, "smsModemSoftwareVersion"
            ),
            "led_dimmer_software_version": IHCSoapClient.__extract_sysinfo(
                xdoc, "ledDimmerSoftwareVersion"
            ),
        }
        return info

    def __extract_sysinfo(xdoc, param) -> str:
        """Internal function to extrach a parameter from system info"""
        element = xdoc.find(
            f"./SOAP-ENV:Body/ns1:getSystemInfo1/ns1:{param}", IHCSoapClient.ihcns
        )
        if element is None:
            return None
        return element.text



"""
Wraps the ihcclient in a more user friendly interface to handle lost connection
Notify thread to handle change notifications
"""

# pylint: disable=invalid-name, bare-except, too-many-instance-attributes
from datetime import datetime, timedelta
import logging
import requests
import threading
import time
from ihcsdk.ihcclient import IHCSoapClient, IHCSTATE_READY

_LOGGER = logging.getLogger(__name__)


class IHCController:
    """
    Implements the notification thread and
    will re-authenticate if needed.
    """

    _mutex = threading.Lock()

    def __init__(self, url: str, username: str, password: str):
        self.client = IHCSoapClient(url)
        self.reauthenticatetimeout = 30
        self.retryinterval = 10
        self._username = username
        self._password = password
        self._ihcevents = {}
        self._ihcvalues = {}
        self._notifythread = threading.Thread(target=self._notify_fn)
        self._notifyrunning = False
        self._newnotifyids = []
        self._project = None

    def is_ihc_controller(url: str) -> bool:
        """Will return True if the url respods like an IHC controller."""
        try:
            client = IHCSoapClient(url)
            response = client.connection.session.get(
                f"{url}/wsdl/controller.wsdl", verify=client.connection.cert_verify()
            )
            if response.status_code != 200:
                return False
            if not response.headers["content-type"].startswith("text/xml"):
                return False
            if response.text.find("getIHCProject") < 0:
                return False
            return True
        except requests.exceptions.RequestException as exp:
            _LOGGER.warning("is_ihc_controller %s", exp)
            return False

    def authenticate(self) -> bool:
        """Authenticate and enable the registered notifications"""
        with IHCController._mutex:
            _LOGGER.debug("Authenticating login on ihc controller")
            if not self.client.authenticate(self._username, self._password):
                _LOGGER.debug("Authentication failed")
                return False
            _LOGGER.debug("Authentication was successful")
            if self._ihcevents:
                self.client.enable_runtime_notifications(self._ihcevents.keys())
            return True

    def disconnect(self):
        """Disconnect by stopping the notification thread
        TODO call disconnect on ihcclient
        """
        self._notifyrunning = False

    def get_runtime_value(self, ihcid: int):
        """Get runtime value with re-authenticate if needed"""
        value = self.client.get_runtime_value(ihcid)
        if value is not None:
            return value
        self.re_authenticate()
        return self.client.get_runtime_value(ihcid)

    def get_runtime_values(self, ihcids):
        """Get runtime value with re-authenticate if needed"""
        value = self.client.get_runtime_values(ihcids)
        if value is not None:
            return value
        self.re_authenticate()
        return self.client.get_runtime_values(ihcids)

    def cycle_bool_value(self, resourceid: int):
        """Turn a booelan resource On and back Off"""
        value = self.client.cycle_bool_value(resourceid)
        if value is not None:
            return value
        self.re_authenticate()
        return self.client.cycle_bool_value(resourceid)

    def set_runtime_value_bool(self, ihcid: int, value: bool) -> bool:
        """Set bool runtime value with re-authenticate if needed"""
        if self.client.set_runtime_value_bool(ihcid, value):
            return True
        self.re_authenticate()
        return self.client.set_runtime_value_bool(ihcid, value)

    def set_runtime_value_int(self, ihcid: int, value: int) -> bool:
        """Set integer runtime value with re-authenticate if needed"""
        if self.client.set_runtime_value_int(ihcid, value):
            return True
        self.re_authenticate()
        return self.client.set_runtime_value_int(ihcid, value)

    def set_runtime_value_float(self, ihcid: int, value: float) -> bool:
        """Set float runtime value with re-authenticate if needed"""
        if self.client.set_runtime_value_float(ihcid, value):
            return True
        self.re_authenticate()
        return self.client.set_runtime_value_float(ihcid, value)

    def set_runtime_value_timer(self, ihcid: int, value: int) -> bool:
        """Set timer runtime value with re-authenticate if needed"""
        if self.client.set_runtime_value_timer(ihcid, value):
            return True
        self.re_authenticate()
        return self.client.set_runtime_value_timer(ihcid, value)

    def set_runtime_value_time(
        self, ihcid: int, hours: int, minutes: int, seconds: int
    ) -> bool:
        """Set time runtime value with re-authenticate if needed"""
        if self.client.set_runtime_value_time(ihcid, hours, minutes):
            return True
        self.re_authenticate()
        return self.client.set_runtime_value_time(ihcid, hours, minutes, seconds)

    def get_project(self, insegments: bool = True) -> str:
        """Get the ihc project and make sure controller is ready before"""
        with IHCController._mutex:
            if self._project is None:
                if self.client.get_state() != IHCSTATE_READY:
                    ready = self.client.wait_for_state_change(IHCSTATE_READY, 10)
                    if ready != IHCSTATE_READY:
                        return None
                if insegments:
                    self._project = self.client.get_project_in_segments()
                else:
                    self._project = self.client.get_project()
        return self._project

    def add_notify_event(self, resourceid: int, callback, delayed=False):
        """Add a notify callback for a specified resource id
        If delayed is set to true the enable request will be send from the
        notofication thread
        """
        with IHCController._mutex:
            if resourceid in self._ihcevents:
                self._ihcevents[resourceid].append(callback)
            else:
                self._ihcevents[resourceid] = [callback]
                if delayed:
                    self._newnotifyids.append(resourceid)
                else:
                    if not self.client.enable_runtime_notification(resourceid):
                        return False
            if not self._notifyrunning:
                self._notifyrunning = True
                self._notifythread.start()

            return True

    def _notify_fn(self):
        """The notify thread function."""
        _LOGGER.debug("Starting notify thread")
        while self._notifyrunning:
            try:
                with IHCController._mutex:
                    # Are there are any new ids to be added?
                    if self._newnotifyids:
                        self.client.enable_runtime_notifications(self._newnotifyids)
                        self._newnotifyids = []

                changes = self.client.wait_for_resource_value_change_list()
                if changes is False:
                    self.re_authenticate(True)
                    continue
                for ihcid, value in changes:
                    if ihcid in self._ihcevents:
                        for callback in self._ihcevents[ihcid]:
                            if (
                                ihcid not in self._ihcvalues
                                or value != self._ihcvalues[ihcid]
                            ):
                                callback(ihcid, value)
                            self._ihcvalues[ihcid] = value
            except Exception as exp:
                _LOGGER.error("Exception in notify thread %s", exp)
                self.re_authenticate(True)

    def re_authenticate(self, notify: bool = False) -> bool:
        """Authenticate again after failure.
        Keep trying with 10 sec interval. If called from the notify thread
        we will not have a timeout, but will end if the notify thread has
        been cancled.
        Will return True if authentication was successful.
        """
        timeout = datetime.now() + timedelta(seconds=self.reauthenticatetimeout)
        while True:
            _LOGGER.debug("Reauthenticating login on ihc controller")
            if self.authenticate():
                return True
            _LOGGER.debug(
                "Authenticate failed - Reauthenticating login on ihc controller in 10 sec"
            )

            # if called from the notify and notify a cancled we do not want to retry
            if notify:
                if not self._notifyrunning:
                    return False
            else:
                if timeout and datetime.now() > timeout:
                    return False
            # wait before we try to authenticate again
            time.sleep(self.retryinterval)



"""Implements soap reqeust using the "requests" module"""

# pylint: disable=too-few-public-methods
import logging
import requests
import xml.etree.ElementTree

from urllib.parse import urlparse
from urllib3.util import Retry
from requests.adapters import HTTPAdapter

_LOGGER = logging.getLogger(__name__)


class IHCConnection(object):
    """Implements a http connection to the controller"""

    soapenvelope = """<?xml version=\"1.0\" encoding=\"UTF-8\"?>
        <s:Envelope xmlns:s=\"http://schemas.xmlsoap.org/soap/envelope/\"
          xmlns:xsd=\"http://www.w3.org/2001/XMLSchema\"
          xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\">
        <s:Body>{body}</s:Body></s:Envelope>"""

    def __init__(self, url: str):
        """Initialize the IHCConnection with a url for the controller"""
        self.url = url
        self.verify = False
        self.last_exception = None
        self.last_response = None
        self.session = requests.Session()
        self.retries = Retry(
            total=3,
            backoff_factor=0.2,
            status_forcelist=[502, 503, 504],
            allowed_methods={"POST"},
        )
        self.session.mount("http://", HTTPAdapter(max_retries=self.retries))

    def cert_verify(self):
        return None

    def soap_action(self, service, action, payloadbody):
        """Do a soap request."""
        payload = self.soapenvelope.format(body=payloadbody).encode("utf-8")
        headers = {
            "Host": urlparse(self.url).netloc,
            "Content-Type": "text/xml; charset=UTF-8",
            "Cache-Control": "no-cache",
            "Content-Length": str(len(payload)),
            "SOAPAction": action,
        }
        try:
            _LOGGER.debug("soap payload %s", payload)
            self.last_exception = None
            response = self.session.post(
                url=self.url + service,
                headers=headers,
                data=payload,
                verify=self.cert_verify(),
            )
            _LOGGER.debug("soap request response status %d", response.status_code)
            if response.status_code != 200:
                self.last_response = response
                return False
            _LOGGER.debug("soap request response %s", response.text)
            xdoc = xml.etree.ElementTree.fromstring(response.text)
            if xdoc is None:
                return False
            return xdoc
        except requests.exceptions.RequestException as exp:
            _LOGGER.error("soap request exception %s", exp)
            self.last_exception = exp
        except xml.etree.ElementTree.ParseError as exp:
            _LOGGER.error("soap request xml parse error %s", exp)
            self.last_exception = exp
            self.last_response = response
        return False



"""Implements soap reqeust using the "requests" module"""

# pylint: disable=too-few-public-methods
import os
import requests

from cryptography.x509 import load_pem_x509_certificate
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from requests.packages.urllib3.util.ssl_ import create_urllib3_context

from ihcsdk.ihcconnection import IHCConnection


class IHCSSLConnection(IHCConnection):
    """Implements a https connection to the controller"""

    def __init__(self, url: str):
        """Initialize the IHCSSLConnection with a url for the controller"""
        super(IHCSSLConnection, self).__init__(url)
        self.cert_file = os.path.dirname(__file__) + "/certs/ihc3.crt"
        self.session.mount(
            "https://",
            CertAdapter(self.get_fingerprint_from_cert(), max_retries=self.retries),
        )

    def get_fingerprint_from_cert(self):
        """Get the fingerprint from the certificate"""
        pem = open(self.cert_file, "rb").read()
        cert = load_pem_x509_certificate(pem, default_backend())
        f = cert.fingerprint(hashes.SHA1())
        return "".join("{:02x}".format(x) for x in f)

    def cert_verify(self):
        return self.cert_file


class CertAdapter(requests.adapters.HTTPAdapter):
    """A adapter for a specific certificate"""

    def __init__(self, fingerprint, **kwargs):
        """Constructor. Store the fingerprint for use when creating the poolmanager."""
        self.fingerprint = fingerprint
        super(CertAdapter, self).__init__(**kwargs)

    def init_poolmanager(self, connections, maxsize, block=False, **pool_kwargs):
        """Create a custom poolmanager"""
        CIPHERS = "DEFAULT:!DH"
        context = create_urllib3_context(ciphers=CIPHERS)
        pool_kwargs["ssl_context"] = context
        pool_kwargs["assert_fingerprint"] = self.fingerprint
        return super(CertAdapter, self).init_poolmanager(
            connections, maxsize, block, **pool_kwargs
        )



"""
Setup of the ihcsdk module
"""

from setuptools import setup

setup(
    name="ihcsdk",
    version="2.8.9",
    description="IHC Python SDK",
    long_description=(
        "SDK for connection to the LK IHC Controller. "
        "Made for interfacing to home assistant"
    ),
    author="Jens Nielsen",
    url="https://github.com/dingusdk/PythonIhcSdk",
    packages=["ihcsdk"],
    install_requires=[
        "requests",
        "cryptography",
    ],
    license="GPL-3.0",
    include_package_data=True,
)
